import pytest

from sqlmodel import Field, SQLModel
from chemql.filters import Filter
from sqlalchemy import and_, or_

filter = Filter()


def show_generated_filter(q):
    print(f"(Generated filter: {q})")


class Table(SQLModel, table=True):
    __tablename__ = "test_table"
    id: int | None = Field(default=None, primary_key=True)
    name: str | None = None


def test_eq():
    q = filter._eq(Table, "name", "test")
    assert str(q) == str(Table.name == "test")
    show_generated_filter(q)


def test_neq():
    q = filter._neq(Table, "name", "test")
    assert str(q) == str(Table.name != "test")
    show_generated_filter(q)


def test_gt():
    q = filter._gt(Table, "id", 1)
    assert str(q) == str(Table.id > 1)
    show_generated_filter(q)


def test_gte():
    q = filter._gte(Table, "id", 1)
    assert str(q) == str(Table.id >= 1)
    show_generated_filter(q)


def test_ilike():
    q = filter._ilike(Table, "name", "test")
    assert str(q) == str(Table.name.ilike("test"))
    show_generated_filter(q)


def test_iregex():
    with pytest.raises(NotImplementedError):
        filter._iregex(Table, "name", "test")


@pytest.mark.parametrize(
    "value, raise_error",
    [(1, True), ("test", True), (["test"], False), ({"test"}, True)],
)
def test_in(value, raise_error):
    if raise_error:
        with pytest.raises(ValueError):
            filter._in(Table, "name", value)
    else:
        assert str(filter._in(Table, "name", value)) == str(Table.name.in_(value))
        show_generated_filter(str(filter._in(Table, "name", value)))


def test_is_null():
    q = filter._is_null(Table, "name", None)
    assert str(q) == str(Table.name == None)
    show_generated_filter(q)

def test_like():
    q = filter._like(Table, "name", "test")
    assert str(q) == str(Table.name.like("test"))
    show_generated_filter(q)

def test_lt():
    q = filter._lt(Table, "id", 1)
    assert str(q) == str(Table.id < 1)
    show_generated_filter(q)

def test_lte():
    q = filter._lte(Table, "id", 1)
    assert str(q) == str(Table.id <= 1)
    show_generated_filter(q)

def test_similar():
    with pytest.raises(NotImplementedError):
        filter._similar(Table, "name", "test")
        

def test_nilike():
    q = filter._nilike(Table, "name", "test")
    assert str(q) == str(~Table.name.ilike("test"))
    show_generated_filter(q)

def test_nin():
    q = filter._nin(Table, "name", ["test"])
    assert str(q) == str(~Table.name.in_(["test"]))
    show_generated_filter(q)

def test_niregex():
    with pytest.raises(NotImplementedError):
        filter._niregex(Table, "name", "test")


def test_nlike():
    q = filter._nlike(Table, "name", "test")
    assert str(q) == str(~Table.name.like("test"))
    show_generated_filter(q)

def test_nregex():
    with pytest.raises(NotImplementedError):
        filter._nregex(Table, "name", "test")


def test_or():
    q = filter._or(Table, [{"name": {"_eq": "test"}}, {"id": {"_eq": 1}}])
    assert str(q) == str(or_(Table.name == "test", Table.id == 1))
    show_generated_filter(q)


def test_and():
    q = filter._and(Table, [{"name": {"_eq": "test"}}, {"id": {"_eq": 1}}])
    assert str(q) == str(and_(Table.name == "test", Table.id == 1))
    show_generated_filter(q)


def test_complex():
    q = filter._and(Table, [{"name": {"_eq": "test"}}, {"id": {"_gt": 1}}])
    assert str(q) == str(and_(Table.name == "test", Table.id > 1))
    show_generated_filter(q)


def test_process_where():
    q = filter.process_where(Table, {"name": {"_eq": "test"}})
    assert str(q) == str(Table.name == "test")
    show_generated_filter(q)


def test_complex_process_where():
    q = filter.process_where(
        Table, {"_and": [{"name": {"_eq": "test"}}, {"id": {"_eq": 1}}]}
    )
    assert str(q) == str(and_(Table.name == "test", Table.id == 1))
    show_generated_filter(q)


def test_complex_process_where_or():
    q = filter.process_where(
        Table, {"_or": [{"name": {"_eq": "test"}}, {"id": {"_eq": 1}}]}
    )
    assert str(q) == str(or_(Table.name == "test", Table.id == 1))
    show_generated_filter(q)


def test_ultra_complex_process_where():
    q = filter.process_where(
        Table,
        {
            "_and": [
                {"_or": [{"name": {"_eq": "test"}}, {"id": {"_eq": 1}}]},
                {"_and": [{"name": {"_eq": "test"}}, {"id": {"_eq": 1}}]},
            ]
        },
    )
    assert str(q) == str(
        and_(
            or_(Table.name == "test", Table.id == 1),
            and_(Table.name == "test", Table.id == 1),
        )
    )
    show_generated_filter(q)